#!/bin/sh
# -*-Scheme-*-
#
# Copyright (C) 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015,
#     2016, 2017, 2018, 2019 Matthew Birkholz
#
# This file is part of a Gtk plugin for MIT/GNU Scheme Pucked.
#
# This plugin is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This plugin is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this plugin; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
# 02110-1301, USA.

# Compile the GTK option.

set -e
: ${MIT_SCHEME_EXE=mit-scheme}
${MIT_SCHEME_EXE} --prepend-library . --batch-mode <<\EOF
(begin

  (parameterize ((param:suppress-loading-message? #t))
    (for-each load-option '(cref cairo ffi)))

  (if (name->package '(GTK))
      (error "The GTK package already exists."))
  (let ((package-set (package-set-pathname "gtk")))
    (if (not (file-modification-time<? "gtk.pkg" package-set))
	(cref/generate-trivial-constructor "gtk"))
    (construct-packages-from-file (fasload package-set)))

  ;; gtk.scm includes the Gtk c-includes, but does not otherwise
  ;; use the FFI.
  (compile-file "gtk" '("gtk-const.bin") (->environment '(gtk)))
  ;; Mostly to set! c-includes:
  (load "gtk" (->environment '(gtk)))

  ;; The wrappers use the FFI, c-includes, and some integrable
  ;; definitions in gtk.scm.  Dependencies between them are
  ;; rare.
  (compile-file "gdk" '("gtk") (->environment '(gtk gdk)))
  (compile-file "gtk-widget" '("gtk") (->environment '(gtk gtk-widget)))
  (compile-file "scm-widget" '("gtk") (->environment '(gtk widget)))
  (compile-file "fix-layout" '("gtk") (->environment '(gtk fix-layout)))
  (compile-file "keys" '("gtk") (->environment '(gtk keys)))
  (compile-file "main" '("gtk") (->environment '(gtk main)))
  (compile-file "gtk-ev" '("gtk") (->environment '(gtk event-viewer)))

  ;; Users of the toolkit interface do NOT use the FFI directly,
  ;; and do not need integrable definitions.
  (compile-file "gtk-graphics" '("gtk") (->environment '(runtime gtk-graphics)))
  (compile-file "fix-demo" '() (->environment '(gtk fix-layout demo)))
  (compile-file "swat" '() (->environment '(gtk swat)))
  (compile-file "swat-pole-zero" '() (->environment '(swat)))

  (cref/generate-constructors "gtk")
  )
EOF
suffix=`echo "(display (microcode-id/operating-system-suffix))" \
	| ${MIT_SCHEME_EXE} --batch-mode`
report=gtk-$suffix.crf
if [ -s "$report" ]; then
    awk "{printf \"$report:%d: %s\\n\",NR,\$0}" $report
fi
