#| -*-Scheme-*-

Copyright (C) 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015,
    2016, 2017, 2018, 2019 Matthew Birkholz

This file is part of a Gtk plugin for MIT/GNU Scheme Pucked.

This plugin is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2 of the License, or (at your
option) any later version.

This plugin is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this plugin; if not, write to the Free Software Foundation,
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

|#

;;;; GtkWidgets/GtkContainers
;;; package: (gtk gtk-widget)

(define-class (<gtk-adjustment> (constructor ())) (<gobject>))

(define-guarantee gtk-adjustment "a <gtk-adjustment>")

(define-integrable-operator (guarantee-live-gtk-adjustment object operator)
  (guarantee-gtk-adjustment object operator)
  (if (not (gobject-live? object))
      (error "Gtk-adjustment dead:" object operator)))

(define (set-gtk-adjustment! adjustment value
			     lower upper page-size
			     step-incr page-incr)
  (assert-glib-locked 'set-gtk-adjustment!)
  (guarantee-live-gtk-adjustment adjustment 'set-gtk-adjustment!)
  (guarantee real? lower 'set-gtk-adjustment!)
  (guarantee real? upper 'set-gtk-adjustment!)
  (guarantee real? page-size 'set-gtk-adjustment!)
  (guarantee real? step-incr 'set-gtk-adjustment!)
  (guarantee real? page-incr 'set-gtk-adjustment!)
  (define-integrable f->e floor->exact)
  (let ((alien (gobject-alien adjustment))
	(new-lower (f->e lower))
	(new-upper (f->e upper))
	(new-value (f->e value))
	(new-page-size (f->e page-size))
	(new-step-incr (f->e step-incr))
	(new-page-incr (f->e page-incr)))
    (let ((old-lower (f->e (C-call "gtk_adjustment_get_lower" alien)))
	  (old-upper (f->e (C-call "gtk_adjustment_get_upper" alien)))
	  (old-value (f->e (C-call "gtk_adjustment_get_value" alien)))
	  (old-page-size (f->e (C-call "gtk_adjustment_get_page_size" alien)))
	  (old-step-incr (f->e (C-call "gtk_adjustment_get_step_increment" alien)))
	  (old-page-incr (f->e (C-call "gtk_adjustment_get_page_increment" alien))))
      (if (not (int:= new-lower old-lower))
	  (C-call "gtk_adjustment_set_lower" alien (->flonum new-lower)))
      (if (not (int:= new-upper old-upper))
	  (C-call "gtk_adjustment_set_upper" alien (->flonum new-upper)))
      (if (not (int:= new-value old-value))
	  (C-call "gtk_adjustment_set_value" alien (->flonum new-value)))
      (if (not (int:= new-page-size old-page-size))
	  (C-call "gtk_adjustment_set_page_size" alien (->flonum
							new-page-size)))
      (if (not (int:= new-step-incr old-step-incr))
	  (C-call "gtk_adjustment_set_step_increment"
		  alien (->flonum new-step-incr)))
      (if (not (int:= new-page-incr old-page-incr))
	  (C-call "gtk_adjustment_set_page_increment"
		  alien (->flonum new-page-incr)))
      ;; gtk_adjustment_changed and gtk_adjustment_value_changed are
      ;; deprecated as of Gtk+ 3.18.  This should now be automatic.
      #;(if (or (not (int:= new-lower old-lower))
	      (not (int:= new-upper old-upper))
	      (not (int:= new-page-size old-page-size))
	      (not (int:= new-step-incr old-step-incr))
	      (not (int:= new-page-incr old-page-incr)))
	  (C-call "gtk_adjustment_changed" alien))
      #;(if (not (int:= new-value old-value))
	  (C-call "gtk_adjustment_value_changed" alien))
      )))

(define (peek-gtk-adjustment adjustment)
  ;; For debugging...
  (let ((alien (gobject-alien adjustment)))
    (list
     'lower (C-call "gtk_adjustment_get_lower" alien)
     'upper (C-call "gtk_adjustment_get_upper" alien)
     'value (C-call "gtk_adjustment_get_value" alien)
     'page-size (C-call "gtk_adjustment_get_page_size" alien)
     'step-increment (C-call "gtk_adjustment_get_step_increment" alien)
     'page-increment (C-call "gtk_adjustment_get_page_increment" alien))))

;;; GtkWidgets

(define-class <gtk-widget> (<gobject>)
  (destroyed? define standard initial-value #f)

  ;; The parent <gtk-widget> or #f.
  (parent define standard initial-value #f))

(define-guarantee gtk-widget "a <gtk-widget>")

;;; This is unfortunate.  We rely on the most specialized method to
;;; call out, creating a specific type of GtkWidget.  We want the
;;; <gobject> method to go first, as usual, to add a glib-cleanup, but
;;; this method to go last, AFTER the most specific (most unusual!)
;;; else it cannot connect its destroy-callback.  To do both would
;;; take... a computed effective method procedure?  For now, rely on
;;; the method that calls out to set-gtk-widget-destroy-callback! as
;;; well as g_object_ref_sink.

#;(define-method initialize-instance ((widget <gtk-widget>))
  (call-next-method widget)
  (g-signal-connect widget (C-callback "destroy") gtk-widget-destroy-callback))

(define (set-gtk-widget-destroy-callback! widget)
  (g-signal-connect widget (C-callback "destroy") gtk-widget-destroy-callback))

;;; Methods of this generic procedure should drop references to other
;;; gobject instances used by @var{widget}.  If these instances are
;;; not shared, they can be explicitly g-object-unref!ed.  Else they
;;; should be dropped, e.g. replaced with #f, to be cleaned up by the
;;; garbage collector.
(define-generic gtk-widget-destroy-callback (widget))

(define (gtk-widget-destroy widget)
  (assert-glib-locked 'gtk-widget-destroy)
  (guarantee-live-gtk-widget widget 'gtk-widget-destroy)
  (C-call "gtk_widget_destroy" (gobject-alien widget)))

(define-integrable-operator (guarantee-live-gtk-widget widget operator)
  (guarantee-gtk-widget widget operator)
  (if (gtk-widget-destroyed? widget)
      (error "Gtk-widget destroyed:" widget operator))
  (if (not (gobject-live? widget))
      (error "Gtk-widget dead:" widget operator)))

(define-method gtk-widget-destroy-callback ((widget <gtk-widget>))
  (if (not (gtk-widget-destroyed? widget))
      (let ((parent (gtk-widget-parent widget)))
	(if (and parent (not (gtk-widget-destroyed? parent)))
	    (container-remove! parent widget))
	(set-gtk-widget-destroyed?! widget #t)
	(gobject-unref! widget))))

(define (gtk-widget-realized? widget)
  (assert-glib-locked 'gtk-widget-realized?)
  (guarantee-live-gtk-widget widget 'gtk-widget-realized?)
  (not (zero? (C-call "gtk_widget_get_realized" (gobject-alien widget)))))

(define (gtk-widget-has-focus? widget)
  (assert-glib-locked 'gtk-widget-has-focus?)
  (guarantee-live-gtk-widget widget 'gtk-widget-has-focus?)
  (not (zero? (C-call "gtk_widget_has_focus" (gobject-alien widget)))))

(define (gtk-widget-drawable? widget)
  (assert-glib-locked 'gtk-widget-drawable?)
  (guarantee-gtk-widget widget 'gtk-widget-drawable?)
  (not (zero? (C-call "gtk_widget_is_drawable" (gobject-alien widget)))))

(define (gtk-widget-grab-focus widget)
  (assert-glib-locked 'gtk-widget-grab-focus)
  (guarantee-gtk-widget widget 'gtk-widget-grab-focus)
  (C-call "gtk_widget_grab_focus" (gobject-alien widget)))

(define (gtk-widget-show widget)
  (assert-glib-locked 'gtk-widget-show)
  (guarantee-gtk-widget widget 'gtk-widget-show)
  (C-call "gtk_widget_show" (gobject-alien widget)))

(define (gtk-widget-show-all widget)
  (assert-glib-locked 'gtk-widget-show-all)
  (guarantee-gtk-widget widget 'gtk-widget-show-all)
  (C-call "gtk_widget_show_all" (gobject-alien widget)))

(define (gtk-widget-error-bell widget)
  (assert-glib-locked 'gtk-widget-error-bell)
  (guarantee-gtk-widget widget 'gtk-widget-error-bell)
  (C-call "gtk_widget_error_bell" (gobject-alien widget)))

(define (gtk-widget-queue-draw widget)
  (assert-glib-locked 'gtk-widget-queue-draw)
  (guarantee-gtk-widget widget 'gtk-widget-queue-draw)
  (C-call "gtk_widget_queue_draw" (gobject-alien widget)))

(define (gtk-widget-queue-resize widget)
  (assert-glib-locked 'gtk-widget-queue-resize)
  (guarantee-gtk-widget widget 'gtk-widget-queue-resize)
  (C-call "gtk_widget_queue_resize" (gobject-alien widget)))

(define (gtk-widget-queue-resize-no-redraw widget)
  (assert-glib-locked 'gtk-widget-queue-resize-no-redraw)
  (guarantee-gtk-widget widget 'gtk-widget-queue-resize-no-redraw)
  (C-call "gtk_widget_queue_resize_no_redraw" (gobject-alien widget)))

(define (gtk-widget-get-pango-context widget)
  (assert-glib-locked 'gtk-widget-get-pango-context)
  (guarantee-gtk-widget widget 'gtk-widget-get-pango-context)
  (C-call "gtk_widget_get_pango_context"
	  (make-alien '|PangoContext|) (gobject-alien widget)))

(define (gtk-widget-create-pango-layout widget #!optional text)
  (assert-glib-locked 'gtk-widget-create-pango-layout)
  (guarantee-gtk-widget widget 'gtk-widget-create-pango-layout)
  (if (not (default-object? text))
      (guarantee string? text 'gtk-widget-create-pango-layout))
  (let* ((layout (make-pango-layout))
	 (alien (gobject-alien layout)))
    (C-call "gtk_widget_create_pango_layout"
	    alien (gobject-alien widget)
	    (if (default-object? text) 0 (string->utf8 text)))
    (error-if-null alien "Could not create:" layout)
    layout))

(define (gtk-widget-get-size widget)
  (assert-glib-locked 'gtk-widget-get-size)
  (let ((alien (gobject-alien widget))
	(allocation (malloc (C-sizeof "GtkAllocation") '|GtkAllocation|)))
    (C-call "gtk_widget_get_allocation" alien allocation)
    (let ((width (C-> allocation "GtkAllocation width"))
	  (height (C-> allocation "GtkAllocation height")))
      (free allocation)
      (cons width height))))

(define (guarantee-size object operator)
  (if (and (fixnum? object) (fix:> object -2))
      object
      (error:wrong-type-argument object "a positive fixnum, 0 or -1" operator)))

(define (gtk-widget-set-size-request widget width height)
  (assert-glib-locked 'gtk-widget-set-size-request)
  (guarantee-gtk-widget widget 'gtk-widget-set-size-request)
  (guarantee-size width 'gtk-widget-set-size-request)
  (guarantee-size height 'gtk-widget-set-size-request)
  (C-call "gtk_widget_set_size_request" (gobject-alien widget) width height))

(define (gtk-widget-set-hexpand widget expand?)
  (assert-glib-locked 'gtk-widget-set-hexpand)
  (C-call "gtk_widget_set_hexpand" (gobject-alien widget) (if expand? 1 0)))

(define (gtk-widget-set-vexpand widget expand?)
  (assert-glib-locked 'gtk-widget-set-vexpand)
  (C-call "gtk_widget_set_vexpand" (gobject-alien widget) (if expand? 1 0)))

(define (set-gtk-widget-size-allocate-callback! widget callback)
  (guarantee-gtk-widget widget 'set-gtk-widget-size-allocate-callback!)
  (guarantee-procedure-of-arity callback 2 'set-gtk-widget-size-allocate-callback!)
  (g-signal-connect widget (C-callback "size_allocate") callback))

(define (set-gtk-widget-realize-callback! widget callback)
  (guarantee-gtk-widget widget 'set-gtk-widget-realize-callback!)
  (guarantee-procedure-of-arity callback 1 'set-gtk-widget-realize-callback!)
  (g-signal-connect widget (C-callback "realize") callback))

(define (set-gtk-widget-unrealize-callback! widget callback)
  (guarantee-gtk-widget widget 'set-gtk-widget-unrealize-callback!)
  (guarantee-procedure-of-arity callback 1 'set-gtk-widget-unrealize-callback!)
  (g-signal-connect widget (C-callback "unrealize") callback))

(define (set-gtk-widget-draw-callback! widget callback)
  (guarantee-gtk-widget widget 'set-gtk-widget-draw-callback!)
  (guarantee-procedure-of-arity callback 2 'set-gtk-widget-draw-callback!)
  (g-signal-connect widget (c-callback "draw") (make-draw-callback callback)))

(define (make-draw-callback callback)
  (named-lambda (draw-callback widget cr)
    (callback widget cr)
    ;; This is ignored in _gtk_widget_draw_internal:
    1))

(define (set-gtk-widget-event-callback! widget callback)
  (guarantee-gtk-widget widget 'set-gtk-widget-event-callback!)
  (guarantee-procedure-of-arity callback 2 'set-gtk-widget-event-callback!)
  (g-signal-connect widget (C-callback "event") (make-event-callback callback)))

(define (make-event-callback callback)
  (named-lambda (event-callback widget event)
    (assert-glib-locked 'event-callback)
    (let ((handled? (callback widget event)))
      (cond ((eq? handled? #t) 1)
	    ((eq? handled? #f) 0)
	    (else (warn "Event callback not boolean:" callback)
		  0)))))

(define (gtk-widget-set-opacity widget opacity)
  (assert-glib-locked 'gtk-widget-set-opacity)
  (guarantee-gtk-widget widget 'gtk-widget-set-opacity)
  (guarantee real? opacity 'gtk-widget-set-opacity)
  (if (not (<= 0. opacity 1.))
      (error:bad-range-argument opacity 'gtk-widget-set-opacity))
  (C-call "gtk_widget_set_opacity" (gobject-alien widget) opacity))

(define (gtk-widget-set-name widget name)
  (assert-glib-locked 'gtk-widget-set-name)
  (guarantee-gtk-widget widget 'gtk-widget-set-name)
  (guarantee string? name 'gtk-widget-set-name)
  (C-call "gtk_widget_set_name" (gobject-alien widget) (string->utf8 name)))

;;; GtkStyleContext & GtkCssProvider

(define-integrable (gtk-widget-style-context widget)
  (let ((style (make-alien '|GtkStyleContext|)))
    (C-call "gtk_widget_get_style_context" style (gobject-alien widget))
    style))

(define-integrable-operator (guarantee-gtk-widget-realized widget operator)
  (guarantee-gtk-widget widget operator)
  (if (not (gtk-widget-realized? widget))
      (warn "Not yet realized:" widget operator)))

(define (gtk-widget-get-style-context widget)
  (guarantee-gtk-widget-realized widget 'gtk-widget-get-style-context)
  (gtk-widget-style-context widget))

(define-integrable (gtk-style-context? object)
  (and (alien? object) (eq? '|GtkStyleContext| (alien/ctype object))))

(define-guarantee gtk-style-context "a GtkStyleContext alien")

(define (gtk-style-context-add-provider style-context css-provider priority)
  (assert-glib-locked 'gtk-style-context-add-provider)
  (guarantee-gtk-style-context style-context 'gtk-style-context-add-provider)
  (guarantee-gtk-css-provider css-provider 'gtk-style-context-add-provider)
  (C-call "gtk_style_context_add_provider"
	  style-context
	  (gobject-alien css-provider)
	  (->gtk-style-provider-priority priority
					 'gtk-style-context-add-provider)))

(define (->gtk-style-provider-priority priority operator)
  (if (exact-nonnegative-integer? priority)
      priority
      (case priority
	((FALLBACK) (C-enum "GTK_STYLE_PROVIDER_PRIORITY_FALLBACK"))
	((THEME) (C-enum "GTK_STYLE_PROVIDER_PRIORITY_THEME"))
	((SETTINGS) (C-enum "GTK_STYLE_PROVIDER_PRIORITY_SETTINGS"))
	((APPLICATION) (C-enum "GTK_STYLE_PROVIDER_PRIORITY_APPLICATION"))
	((USER) (C-enum "GTK_STYLE_PROVIDER_PRIORITY_USER"))
	(else (error:wrong-type-argument
	       priority "a GtkStylProvider priority" operator)))))

(define-class <gobject-with-gerror*> (<gobject>)
  ;; A <gobject> with associated *GError that gets freed when the
  ;; <gobject> is GCed.
  (gerror* define accessor accessor gobject-gerror*
	   initializer (lambda () (make-alien '(* |GError|)))))

(define-method initialize-instance ((object <gobject-with-gerror*>))
  (call-next-method object)
  (assert-glib-locked '(initialize-instance <gobject-with-gerror*>))
  (let ((gerror* (gobject-gerror* object)))
    (C-call "g_try_malloc0" gerror* (C-sizeof "* GError"))
    (error-if-null gerror* "Could not allocate:" gerror*)
    (add-glib-cleanup object (make-gerror*-cleanup gerror*))))

(define (make-gerror*-cleanup gerror*)
  (named-lambda (gerror*-cleanup)
    (assert-glib-locked 'gerror*-cleanup)
    (assert-without-interruption 'gerror*-cleanup)
    (if (not (alien-null? gerror*))
	(let ((gerror (make-alien '|GError|)))
	  (C-> gerror* "* GError" gerror)
	  (if (not (alien-null? gerror))
	      (C-call "g_error_free" gerror))
	  (C-call "g_free" gerror*)
	  (alien-null! gerror*)))))

(define (error-if-gerror* gerror* message . data)
  (assert-glib-locked 'error-if-gerror*)
  (let ((gerror (C-> gerror* "* GError")))
    (if (not (alien-null? gerror))
	(let ((errmsg (c-peek-cstring (C-> gerror "GError message"))))
	  (without-interruption
	   (lambda ()
	     (C->= gerror* "* GError" 0)
	     (C-call "g_error_free" gerror)))
	  (apply error message errmsg data)))))

(define-class (<gtk-css-provider> (constructor ()))
    (<gobject-with-gerror*>))

(define-guarantee gtk-css-provider "a <gtk-css-provider>")

(define (gtk-css-provider-new)
  (assert-glib-locked 'gtk-css-provider-new)
  (let* ((object (make-gtk-css-provider))
	 (alien (gobject-alien object)))
    (C-call "gtk_css_provider_new" alien)
    (error-if-null alien "Could not create:" object)
    (C-call "g_object_ref_sink" alien alien)
    object))

(define (gtk-css-provider-get-default)
  (assert-glib-locked 'gtk-css-provider-get-default)
  (let* ((object (make-gtk-css-provider))
	 (alien (gobject-alien object)))
    (C-call "gtk_css_provider_get_default" alien)
    (error-if-null alien "Could not get default GtkCssProvider:" object)
    (C-call "g_object_ref" alien alien)
    object))

(define (gtk-css-provider-get-named name variant)
  (assert-glib-locked 'gtk-css-provider-get-named)
  (guarantee string? name 'gtk-css-provider-get-named)
  (let* ((v (if (eq? #f variant)
		0
		(begin
		  (guarantee string? variant 'gtk-css-provider-get-named)
		  (string->utf8 variant))))
	 (object (make-gtk-css-provider))
	 (alien (gobject-alien object)))
    (C-call "gtk_css_provider_get_named" alien (string->utf8 name) v)
    (error-if-null alien "Could not get named GtkCssProvider:"
		   object name variant)
    (C-call "g_object_ref" alien alien)
    object))

(define (gtk-css-provider-load-from-data provider string)
  (assert-glib-locked 'gtk-css-provider-load-from-data)
  (guarantee-gtk-css-provider provider 'gtk-css-provider-load-from-data)
  (guarantee string? string 'gtk-css-provider-load-from-data)
  (let ((alien (gobject-alien provider))
	(gerror* (gobject-gerror* provider))
	(string-bv (string->utf8 string)))
    (C-call "gtk_css_provider_load_from_data" alien string-bv -1 gerror*)
    (error-if-gerror* gerror* "Could not load GtkCssProvider data:" provider)))

(define-guarantee gfile "a <gfile>")

(define (gtk-css-provider-load-from-file provider gfile)
  (assert-glib-locked 'gtk-css-provider-load-from-file)
  (guarantee-gtk-css-provider provider 'gtk-css-provider-load-from-file)
  (guarantee-gfile gfile 'gtk-css-provider-load-from-file)
  (let ((alien (gobject-alien provider))
	(gerror* (gobject-gerror* provider))
	(gfile-alien (gobject-alien gfile)))
    (C-call "gtk_css_provider_load_from_file" alien gfile-alien gerror*)
    (error-if-gerror* gerror* "Could not load GtkCssProvider GFile:"
		      provider gfile)))

(define (gtk-css-provider-load-from-path provider pathname)
  (assert-glib-locked 'gtk-css-provider-load-from-path)
  (guarantee-gtk-css-provider provider 'gtk-css-provider-load-from-file)
  (let ((namestring (->namestring
		     (pathname-simplify
		      (merge-pathnames pathname (working-directory-pathname)))))
	(alien (gobject-alien provider))
	(gerror* (gobject-gerror* provider)))
    (C-call "gtk_css_provider_load_from_path" alien
	    (string->utf8 namestring) gerror*)
    (error-if-gerror* gerror* "Could not load GtkCssProvider path:"
		      provider namestring)))

;;; GtkContainers

(define-class <gtk-container> (<gtk-widget>)
  ;; A list of child gtk-widgets, in the reverse of the order in which
  ;; they were added.
  (reverse-children define standard initial-value '()))

(define-method gtk-widget-destroy-callback ((container <gtk-container>))
  (call-next-method container)
  (for-each gtk-widget-destroy (gtk-container-reverse-children container)))

(define-guarantee gtk-container "a <gtk-container>")

(define (gtk-container-children container)
  (reverse (gtk-container-reverse-children container)))

(define (gtk-bin-child container)
  (let ((c (gtk-container-reverse-children container)))
    (if (pair? c) (car c) #f)))

(define (gtk-container-add parent child)
  (assert-glib-locked 'gtk-container-add)
  (guarantee-gtk-container parent 'gtk-container-add)
  (guarantee-gtk-widget child 'gtk-container-add)
  (container-add! parent child)
  (C-call "gtk_container_add" (gobject-alien parent) (gobject-alien child)))

(define (gtk-container-remove parent child)
  (assert-glib-locked 'gtk-container-remove)
  (guarantee-gtk-container parent 'gtk-container-remove)
  (guarantee-gtk-widget child 'gtk-container-remove)
  (container-remove! parent child)
  (C-call "gtk_container_remove" (gobject-alien parent) (gobject-alien child)))

(define (gtk-container-set-border-width container width)
  (assert-glib-locked 'gtk-container-set-border-width)
  (guarantee-gtk-container container 'gtk-container-set-border-width)
  (guarantee positive-fixnum? width 'gtk-container-set-border-width)
  (C-call "gtk_container_set_border_width" (gobject-alien container) width))

(define (container-add! container child)
  (without-interruption
   (lambda ()
     (if (gtk-widget-parent child)
	 (error "Already a child:" child))
     (let ((children (gtk-container-reverse-children container)))
       (if (memq child children)
	   (error "Already a child:" child container))
       (set-gtk-container-reverse-children! container (cons child children)))
     (set-gtk-widget-parent! child container))))

(define (container-remove! container child)
  (without-interruption
   (lambda ()
     (if (not (eq? container (gtk-widget-parent child)))
	 (error "Not parent:" container child))
     (let ((children (gtk-container-reverse-children container)))
       (if (memq child children)
	   (set-gtk-container-reverse-children! container (delq! child children))
	   (error "Not in container:" child container)))
     (set-gtk-widget-parent! child #f))))

;;; GtkLabels

(define-class (<gtk-label> (constructor () (string))) (<gtk-widget>))

(define-guarantee gtk-label "a <gtk-label>")

(define-method initialize-instance ((label <gtk-label>) string)
  (call-next-method label)
  (assert-glib-locked '(initialize-instance <gtk-label>))
  (let ((alien (gobject-alien label)))
    (C-call "gtk_label_new" alien (string->utf8 string))
    (error-if-null alien "Could not create:" label string)
    (C-call "g_object_ref_sink" alien alien))
  (set-gtk-widget-destroy-callback! label))

(define (gtk-label-new string)
  (guarantee string? string 'gtk-label-new)
  (make-gtk-label string))

(define (gtk-label-get-text label)
  (assert-glib-locked 'gtk-label-get-text)
  (guarantee-gtk-label label 'gtk-label-get-text)
  (let ((retval (make-alien '|gchar|)))
    (C-call "gtk_label_get_text" retval (gobject-alien label))
    (c-peek-cstring retval)))

(define (gtk-label-set-text label string)
  (assert-glib-locked 'gtk-label-set-text)
  (guarantee-gtk-label label 'gtk-label-set-text)
  (guarantee string? string 'gtk-label-set-text)
  (C-call "gtk_label_set_text" (gobject-alien label) (string->utf8 string)))

(define (gtk-label-set-width-chars label n-chars)
  (assert-glib-locked 'gtk-label-set-width-chars)
  (guarantee non-negative-fixnum? n-chars 'set-label-width!)
  (C-call "gtk_label_set_width_chars" (gobject-alien label) n-chars))

;;; GtkButtons

(define-class (<gtk-button> (constructor gtk-button-new ())) (<gtk-container>))

(define-guarantee gtk-button "a <gtk-button>")

(define-method initialize-instance ((button <gtk-button>))
  (call-next-method button)
  (assert-glib-locked '(initialize-instance <gtk-button>))
  (let ((alien (gobject-alien button)))
    (C-call "gtk_button_new" alien)
    (error-if-null alien "Could not create:" button)
    (C-call "g_object_ref_sink" alien alien))
  (set-gtk-widget-destroy-callback! button))

(define (set-gtk-button-clicked-callback! button callback)
  (guarantee-gtk-button button 'set-gtk-button-clicked-callback!)
  (guarantee-procedure-of-arity callback 1 'set-gtk-button-clicked-callback!)
  (g-signal-connect button (C-callback "clicked")
		    (make-clicked-callback callback)))

(define (make-clicked-callback callback)
  (named-lambda (clicked-callback button)
    (callback button)))

(define-class (<gtk-check-button> (constructor gtk-check-button-new ()))
    (<gtk-container>))

(define-guarantee gtk-check-button "a <gtk-check-button>")

(define-method initialize-instance ((button <gtk-check-button>))
  (call-next-method button)
  (assert-glib-locked '(initialize-instance <gtk-check-button>))
  (let ((alien (gobject-alien button)))
    (C-call "gtk_check_button_new" alien)
    (error-if-null alien "Could not create:" button)
    (C-call "g_object_ref_sink" alien alien))
  (set-gtk-widget-destroy-callback! button))

(define (gtk-check-button-get-active button)
  (assert-glib-locked 'gtk-check-button-get-active)
  (guarantee-gtk-check-button button 'gtk-check-button-get-active)
  (not (fix:zero?
	(C-call "gtk_toggle_button_get_active" (gobject-alien button)))))

(define (gtk-check-button-set-active button active?)
  (assert-glib-locked 'gtk-check-button-set-active)
  (guarantee-gtk-check-button button 'gtk-check-button-set-active)
  (C-call "gtk_toggle_button_set_active"
	  (gobject-alien button) (if active? 1 0)))

(define (set-gtk-check-button-toggled-callback! button callback)
  (guarantee-gtk-check-button button 'set-gtk-check-button-toggled-callback!)
  (guarantee-procedure-of-arity callback 1 'set-gtk-check-button-toggled-callback!)
  (g-signal-connect button (C-callback "toggled")
		    (make-toggled-callback callback)))

(define (make-toggled-callback callback)
  (named-lambda (gtk-check-button-toggled-callback button)
     (callback button)))

;;; GtkGrids

(define-class (<gtk-grid> (constructor gtk-grid-new ()))
    (<gtk-container>))

(define-guarantee gtk-grid "a <gtk-grid>")

(define-method initialize-instance ((grid <gtk-grid>))
  (call-next-method grid)
  (assert-glib-locked '(initialize-instance <gtk-grid>))
  (let ((alien (gobject-alien grid)))
    (C-call "gtk_grid_new" alien)
    (error-if-null alien "Could not create:" grid)
    (C-call "g_object_ref_sink" alien alien))
  (set-gtk-widget-destroy-callback! grid))

(define (gtk-grid-set-row-homogeneous grid homogeneous?)
  (assert-glib-locked 'gtk-grid-set-row-homogeneous)
  (guarantee-gtk-grid grid 'gtk-grid-set-row-homogeneous)
  (C-call "gtk_grid_set_row_homogeneous" (gobject-alien grid)
	  (if homogeneous? 1 0)))

(define (gtk-grid-set-column-homogeneous grid homogeneous?)
  (assert-glib-locked 'gtk-grid-set-column-homogeneous)
  (guarantee-gtk-grid grid 'gtk-grid-set-column-homogeneous)
  (C-call "gtk_grid_set_column_homogeneous" (gobject-alien grid)
	  (if homogeneous? 1 0)))

(define (gtk-grid-set-row-spacing grid spacing)
  (assert-glib-locked 'gtk-grid-set-row-spacing)
  (guarantee-gtk-grid grid 'gtk-grid-set-row-spacing)
  (guarantee non-negative-fixnum? spacing 'gtk-grid-set-row-spacing)
  (C-call "gtk_grid_set_row_spacing" (gobject-alien grid) spacing))

(define (gtk-grid-set-column-spacing grid spacing)
  (assert-glib-locked 'gtk-grid-set-column-spacing)
  (guarantee-gtk-grid grid 'gtk-grid-set-column-spacing)
  (guarantee non-negative-fixnum? spacing 'gtk-grid-set-column-spacing)
  (C-call "gtk_grid_set_column_spacing" (gobject-alien grid) spacing))

(define (gtk-grid-attach grid widget left top width height)
  (assert-glib-locked 'gtk-grid-attach)
  (guarantee-gtk-grid grid 'gtk-grid-attach)
  (guarantee-gtk-widget widget 'gtk-grid-attach)
  (guarantee fixnum? left 'gtk-grid-attach)
  (guarantee fixnum? top 'gtk-grid-attach)
  (guarantee fixnum? width 'gtk-grid-attach)
  (guarantee fixnum? height 'gtk-grid-attach)
  (container-add! grid widget)
  (C-call "gtk_grid_attach" (gobject-alien grid) (gobject-alien widget)
	  left top width height))

(define (gtk-grid-attach-next-to grid child sibling side width height)
  (assert-glib-locked 'gtk-grid-attach-next-to)
  (guarantee-gtk-grid grid 'gtk-grid-attach-next-to)
  (guarantee-gtk-widget child 'gtk-grid-attach-next-to)
  (if sibling (guarantee-gtk-widget sibling 'gtk-grid-attach-next-to))
  (guarantee fixnum? width 'gtk-grid-attach-next-to)
  (guarantee fixnum? height 'gtk-grid-attach-next-to)
  (let ((side-num (->side side 'gtk-grid-attach-next-to)))
    (container-add! grid child)
    (C-call "gtk_grid_attach_next_to"
	    (gobject-alien grid)
	    (gobject-alien child)
	    (if sibling (gobject-alien sibling) 0)
	    side-num
	    width height)))

(define (->side object operator)
  (case object
    ((left) (C-enum "GTK_POS_LEFT"))
    ((right) (C-enum "GTK_POS_RIGHT"))
    ((top) (C-enum "GTK_POS_TOP"))
    ((bottom) (C-enum "GTK_POS_BOTTOM"))
    (else (error:wrong-type-argument object "a GtkPositionType" operator))))

(define (gtk-orientable-get-orientation orientable)
  (assert-glib-locked 'gtk-orientable-get-orientation)
  (let ((o (C-call "gtk_orientable_get_orientation"
		   (gobject-alien orientable))))
    (cond ((int:= o (C-enum "GTK_ORIENTATION_VERTICAL"))
	   'VERTICAL)
	  ((int:= o (C-enum  "GTK_ORIENTATION_HORIZONTAL"))
	   'HORIZONTAL)
	  (else (error "Unexpected orientation:" o)))))

(define (->gtk-orientation orientation operator)
  (case orientation
    ((VERTICAL) (C-enum "GTK_ORIENTATION_VERTICAL"))
    ((HORIZONTAL) (C-enum "GTK_ORIENTATION_HORIZONTAL"))
    (else (error:wrong-type-argument
	   orientation
	   "an orientation (vertical or horizontal)"
	   operator))))

(define (gtk-orientable-set-orientation orientable orientation)
  (assert-glib-locked 'gtk-orientable-set-orientation)
  (C-call "gtk_orientable_set_orientation" (gobject-alien orientable)
	  (->gtk-orientation orientation 'gtk-orientable-set-orientation)))

;;; GtkFrames

(define-class (<gtk-frame> (constructor () (label))) (<gtk-container>))

(define-guarantee gtk-frame "a <gtk-frame>")

(define-method initialize-instance ((frame <gtk-frame>) label)
  (call-next-method frame)
  (assert-glib-locked '(initialize-instance <gtk-frame>))
  (let ((alien (gobject-alien frame)))
    (C-call "gtk_frame_new" alien
	    (if (string-null? label) 0 (string->utf8 label)))
    (error-if-null alien "Could not create:" frame)
    (C-call "g_object_ref_sink" alien alien))
  (set-gtk-widget-destroy-callback! frame))

(define (gtk-frame-new label)
  (guarantee string? label 'gtk-frame-new)
  (make-gtk-frame label))

(define (gtk-frame-set-shadow-type frame type)
  (assert-glib-locked 'gtk-frame-set-shadow-type)
  (let ((t (->gtk-shadow-type type 'gtk-frame-set-shadow-type)))
    (C-call "gtk_frame_set_shadow_type" (gobject-alien frame) t)))

(define (->gtk-shadow-type object operator)
  (case object
    ((none) (C-enum "GTK_SHADOW_NONE"))
    ((in) (C-enum "GTK_SHADOW_IN"))
    ((out) (C-enum "GTK_SHADOW_OUT"))
    ((etched-in) (C-enum "GTK_SHADOW_ETCHED_IN"))
    ((etched-out) (C-enum "GTK_SHADOW_ETCHED_OUT"))
    (else
     (error:wrong-type-argument object "a GtkShadowType symbol" operator))))

;;; GtkScrolledWindows

(define-class (<gtk-scrolled-window> (constructor ()))
  (<gtk-container>))

(define-guarantee gtk-scrolled-window "a <gtk-scrolled-window>")

(define (gtk-scrolled-window-new)
  (assert-glib-locked 'gtk-scrolled-window-new)
  (let* ((window (make-gtk-scrolled-window))
	 (alien (gobject-alien window)))
    (C-call "gtk_scrolled_window_new" alien 0 0)
    (error-if-null alien "Could not create:" window)
    (C-call "g_object_ref_sink" alien alien)
    (set-gtk-widget-destroy-callback! window)
    window))

(define (gtk-scrolled-window-set-policy window horizontal vertical)
  (assert-glib-locked 'gtk-scrolled-window-set-policy)
  (guarantee-gtk-scrolled-window window 'gtk-scrolled-window-set-policy)
  (C-call "gtk_scrolled_window_set_policy" (gobject-alien window)
	  (->policy horizontal 'gtk-scrolled-window-set-policy)
	  (->policy vertical 'gtk-scrolled-window-set-policy)))

(define (gtk-scrolled-window-set-placement window placement)
  (assert-glib-locked 'gtk-scrolled-window-set-placement)
  (guarantee-gtk-scrolled-window window 'gtk-scrolled-window-set-placement)
  (C-call "gtk_scrolled_window_set_placement" (gobject-alien window)
	  (->placement placement 'gtk-scrolled-window-set-placement)))

(define (->policy object operator)
  (case object
    ((ALWAYS)	(C-enum "GTK_POLICY_ALWAYS"))
    ((AUTO)	(C-enum "GTK_POLICY_AUTOMATIC"))
    ((NEVER)	(C-enum "GTK_POLICY_NEVER"))
    (else (error:wrong-type-argument
	   object "a symbol -- one of ALWAYS, AUTO or NEVER" operator))))

(define (->placement object operator)
  (case object
    ((TOP-LEFT) (C-enum "GTK_CORNER_TOP_LEFT"))
    ((BOTTOM-LEFT) (C-enum "GTK_CORNER_BOTTOM_LEFT"))
    ((TOP-RIGHT) (C-enum "GTK_CORNER_TOP_RIGHT"))
    ((BOTTOM-RIGHT) (C-enum "GTK_CORNER_BOTTOM_RIGHT"))
    (else (error:wrong-type-argument
	   object
	   "a symbol -- one of TOP-LEFT, BOTTOM-LEFT, TOP-RIGHT or BOTTOM-RIGHT"
	   operator))))

(define-class (<gtk-scrolled-view> (constructor ()))
  (<gtk-scrolled-window>))

(define-guarantee gtk-scrolled-view "a <gtk-scrolled-view>")

(define (gtk-scrolled-view-new)
  (assert-glib-locked 'gtk-scrolled-view-new)
  (let* ((window (make-gtk-scrolled-view))
	 (alien (gobject-alien window)))
    (C-call "gtk_scrolled_view_new" alien 0 0)
    (error-if-null alien "Could not create:" window)
    (C-call "g_object_ref_sink" alien alien)
    (set-gtk-widget-destroy-callback! window)
    window))

;;; GtkPaneds

(define-class (<gtk-paned> (constructor ()))
    (<gtk-container>)
  (child1 define standard accessor gtk-paned-get-child1 initial-value #f)
  (child2 define standard accessor gtk-paned-get-child2 initial-value #f))

(define (gtk-paned-new orientation)
  (assert-glib-locked 'gtk-paned-new)
  (let ((orient (->gtk-orientation orientation 'gtk-paned-new))
	(paned (make-gtk-paned)))
    (let ((alien (gobject-alien paned)))
      (C-call "gtk_paned_new" alien orient)
      (error-if-null alien "Could not create:" paned orient)
      (C-call "g_object_ref_sink" alien alien))
    (set-gtk-widget-destroy-callback! paned)
    paned))

(define (gtk-paned-pack1 paned child1 resize? shrink?)
  (assert-glib-locked 'gtk-paned-pack1)
  (guarantee-gtk-widget child1 'gtk-paned-pack1)
  (let ((existing (gtk-paned-get-child1 paned)))
    (if (and existing (not (gtk-widget-destroyed? existing)))
	(container-remove! paned existing)))
  (set-gtk-paned-child1! paned child1)
  (container-add! paned child1)
  (C-call "gtk_paned_pack1" (gobject-alien paned) (gobject-alien child1)
	  (if resize? 1 0) (if shrink? 1 0)))

(define (gtk-paned-pack2 paned child2 resize? shrink?)
  (assert-glib-locked 'gtk-paned-pack2)
  (guarantee-gtk-widget child2 'gtk-paned-pack2)
  (let ((existing (gtk-paned-get-child2 paned)))
    (if (and existing (not (gtk-widget-destroyed? existing)))
	(container-remove! paned existing)))
  (set-gtk-paned-child2! paned child2)
  (container-add! paned child2)
  (C-call "gtk_paned_pack2" (gobject-alien paned) (gobject-alien child2)
	  (if resize? 1 0) (if shrink? 1 0)))

(define (gtk-paned-get-position paned)
  (assert-glib-locked 'gtk-paned-get-position)
  (C-call "gtk_paned_get_position" (gobject-alien paned)))

(define (gtk-paned-set-position paned child1-size)
  (assert-glib-locked 'gtk-paned-set-position)
  (C-call "gtk_paned_set_position" (gobject-alien paned) child1-size))

(define-class (<gtk-paned-view> (constructor ()))
    (<gtk-paned>))

(define (gtk-paned-view-new orientation)
  (let ((paned (make-gtk-paned-view)))
    (gtk-paned-view-init paned orientation)
    (set-gtk-widget-destroy-callback! paned)
    paned))

(define (gtk-paned-view-init paned orientation)
  ;; Used by Edwin's gtk-vpaned-new and gtk-hpaned-new.
  (assert-glib-locked 'gtk-paned-view-init)
  (let ((orient (->gtk-orientation orientation 'gtk-paned-view-init))
	(alien (gobject-alien paned)))
    (C-call "gtk_paned_view_new" alien orient)
    (error-if-null alien "Could not create:" paned)
    (C-call "g_object_ref_sink" alien alien)))

;;; GtkWindows

(define-class (<gtk-window> (constructor gtk-window-new () (type)))
	      (<gtk-container>)
  ;; 'POPUP or 'TOPLEVEL
  (type define accessor))

(define-guarantee gtk-window "a <gtk-window>")

(define-method initialize-instance ((window <gtk-window>) type)
  (call-next-method window)
  (assert-glib-locked '(initialize-instance <gtk-window>))
  (let ((type (->window-type type 'gtk-window-new))
	(alien (gobject-alien window)))
    (C-call "gtk_window_new" alien type)
    (error-if-null alien "Could not create:" window type)
    (C-call "g_object_ref_sink" alien alien)
    (set-gtk-widget-destroy-callback! window)
    (set! toplevel-windows (cons window toplevel-windows))))

(define toplevel-windows '())

(define-method gtk-widget-destroy-callback ((window <gtk-window>))
  (call-next-method window)
  (set! toplevel-windows (delq! window toplevel-windows)))

(define (->window-type type operator)
  (case type
    ((TOPLEVEL) (C-enum "GTK_WINDOW_TOPLEVEL"))
    ((POPUP) (C-enum "GTK_WINDOW_POPUP"))
    (else
     (error:wrong-type-argument
      type "a symbol -- one of TOPLEVEL or POPUP" operator))))

(define (gtk-window-set-accept-focus window accept?)
  (assert-glib-locked 'gtk-window-set-accept-focus)
  (guarantee-gtk-window window 'gtk-window-set-accept-focus)
  (C-call "gtk_window_set_accept_focus"
	  (gobject-alien window)
	  (if (eq? accept? #f) 0 1)))

(define (gtk-window-set-title window title)
  (assert-glib-locked 'gtk-window-set-title)
  (guarantee-gtk-window window 'gtk-window-set-title)
  (guarantee string? title 'gtk-window-set-title)
  (C-call "gtk_window_set_title" (gobject-alien window) (string->utf8 title)))

(define (gtk-window-set-type-hint window hint)
  (assert-glib-locked 'gtk-window-set-type-hint)
  (guarantee-gtk-window window 'gtk-window-set-type-hint)
  (let ((type-hint (->type-hint hint 'gtk-window-set-type-hint)))
    (C-call "gtk_window_set_type_hint" (gobject-alien window) type-hint)))

(define (->type-hint object operator)
  (case object
    ((NORMAL) (C-enum "GDK_WINDOW_TYPE_HINT_NORMAL"))
    ((DIALOG) (C-enum "GDK_WINDOW_TYPE_HINT_DIALOG"))
    ((MENU) (C-enum "GDK_WINDOW_TYPE_HINT_MENU"))
    ((TOOLBAR) (C-enum "GDK_WINDOW_TYPE_HINT_TOOLBAR"))
    ((SPLASHSCREEN) (C-enum "GDK_WINDOW_TYPE_HINT_SPLASHSCREEN"))
    ((UTILITY) (C-enum "GDK_WINDOW_TYPE_HINT_UTILITY"))
    ((DOCK) (C-enum "GDK_WINDOW_TYPE_HINT_DOCK"))
    ((DESKTOP) (C-enum "GDK_WINDOW_TYPE_HINT_DESKTOP"))
    ((DROPDOWN-MENU) (C-enum "GDK_WINDOW_TYPE_HINT_DROPDOWN_MENU"))
    ((POPUP-MENU) (C-enum "GDK_WINDOW_TYPE_HINT_POPUP_MENU"))
    ((TOOLTIP) (C-enum "GDK_WINDOW_TYPE_HINT_TOOLTIP"))
    ((NOTIFICATION) (C-enum "GDK_WINDOW_TYPE_HINT_NOTIFICATION"))
    ((COMBO) (C-enum "GDK_WINDOW_TYPE_HINT_COMBO"))
    ((DND) (C-enum "GDK_WINDOW_TYPE_HINT_DND"))
    (else (error:wrong-type-argument object "a GdkWindow type hint" operator))))

(define (gtk-window-get-default-size window receiver)
  (assert-glib-locked 'gtk-window-get-default-size)
  (guarantee-gtk-window window 'gtk-window-get-default-size)
  (let* ((*width (malloc (fix:* 2 (C-sizeof "gint")) 'gint))
	 (*height (alien-byte-increment *width (C-sizeof "gint") 'gint)))
    (C-call "gtk_window_get_default_size"
	    (gobject-alien window) *width *height)
    (let ((width (C-> *width "gint"))
	  (height (C-> *height "gint")))
      (free *width)
      (receiver width height))))

(define (gtk-window-set-default-size window width height)
  (assert-glib-locked 'gtk-window-set-default-size)
  (guarantee-gtk-window window 'gtk-window-set-default-size)
  (guarantee integer? width 'gtk-window-set-default-size)
  (guarantee integer? height 'gtk-window-set-default-size)
  (C-call "gtk_window_set_default_size" (gobject-alien window) width height))

(define (gtk-window-set-geometry-hints window widget . hints)
  (assert-glib-locked 'gtk-window-set-geometry-hints)
  (guarantee-gtk-window window 'gtk-window-set-geometry-hints)
  (guarantee-gtk-widget widget 'gtk-window-set-geometry-hints)
  (guarantee-restricted-keyword-list
   hints '(min-width min-height max-width max-height base-width base-height
		     width-increment height-increment min-aspect max-aspect
		     gravity) 'gtk-window-set-geometry-hints)
  (let ((geometry (malloc (C-sizeof "GdkGeometry") '|GdkGeometry|))
	(mask 0))

    (define-integrable (value key)
      (get-keyword-value hints key))

    (define-integrable-operator (get-fixnum name)
      (let ((v (value name)))
	(cond ((default-object? v) #f)
	      ((not (fixnum? v)) (error "Not a fixnum:" v name 'gtk-window-set-geometry-hints))
	      ((fix:< v -1) (error "Negative:" v name 'gtk-window-set-geometry-hints))
	      (else v))))

    (define-integrable-operator (get-real name)
      (let ((v (value name)))
	(cond ((default-object? v) #f)
	      ((not (real? v)) (error "Not real:" v name 'gtk-window-set-geometry-hints))
	      (else v))))

    (define-integrable (get-gravity)
      (let ((v (value 'gravity)))
	(if (default-object? v)
	    #f
	    (case v
	      ((north)		(C-enum "GDK_GRAVITY_NORTH"))
	      ((northeast)	(C-enum "GDK_GRAVITY_NORTH_EAST"))
	      ((east)		(C-enum "GDK_GRAVITY_EAST"))
	      ((southeast)	(C-enum "GDK_GRAVITY_SOUTH_EAST"))
	      ((south)		(C-enum "GDK_GRAVITY_SOUTH"))
	      ((southwest)	(C-enum "GDK_GRAVITY_SOUTH_WEST"))
	      ((west)		(C-enum "GDK_GRAVITY_WEST"))
	      ((northwest)	(C-enum "GDK_GRAVITY_NORTH_WEST"))
	      ((center)		(C-enum "GDK_GRAVITY_CENTER"))
	      ((static)		(C-enum "GDK_GRAVITY_STATIC"))
	      (else (error "Not a gravity:" v 'gtk-window-set-geometry-hints))))))

    (let ((width (get-fixnum 'min-width))
	  (height (get-fixnum 'min-height)))
      (cond ((and width height)
	     (C->= geometry "GdkGeometry min_width" width)
	     (C->= geometry "GdkGeometry min_height" height)
	     (set! mask (fix:+ mask (C-enum "GDK_HINT_MIN_SIZE"))))
	    ((and (not width) (not height)))
	    (else (error "Both min-width and min-height are required:" hints))))

    (let ((width (get-fixnum 'max-width))
	  (height (get-fixnum 'max-height)))
      (cond ((and width height)
	     (C->= geometry "GdkGeometry max_width" width)
	     (C->= geometry "GdkGeometry max_height" height)
	     (set! mask (fix:+ mask (C-enum "GDK_HINT_MAX_SIZE"))))
	    ((and (not width) (not height)))
	    (else (error "Both max-width and max-height are required:" hints))))

    (let ((width (get-fixnum 'base-width))
	  (height (get-fixnum 'base-height)))
      (cond ((and width height)
	     (C->= geometry "GdkGeometry base_width" width)
	     (C->= geometry "GdkGeometry base_height" height)
	     (set! mask (fix:+ mask (C-enum "GDK_HINT_BASE_SIZE"))))
	    ((and (not width) (not height)))
	    (else (error "Both base-width and base-height are required:" hints))))

    (let ((width (get-fixnum 'width-increment))
	  (height (get-fixnum 'height-increment)))
      (cond ((and width height)
	     (C->= geometry "GdkGeometry width_inc" width)
	     (C->= geometry "GdkGeometry height_inc" height)
	     (set! mask (fix:+ mask (C-enum "GDK_HINT_RESIZE_INC"))))
	    ((and (not width) (not height)))
	    (else (error "Both width-increment and height-increment are required:" hints))))

    (let ((min (get-real 'min-aspect))
	  (max (get-real 'max-aspect)))
      (cond ((and min max)
	     (C->= geometry "GdkGeometry min_aspect" min)
	     (C->= geometry "GdkGeometry max_aspect" max)
	     (set! mask (fix:+ mask (C-enum "GDK_HINT_ASPECT"))))
	    ((and (not min) (not max)))
	    (else (error "Both min-aspect and max-aspect must be specified:" hints))))

    (let ((gravity (get-gravity)))
      (if gravity
	  (begin
	    (C->= geometry "GdkGeometry win_gravity" gravity)
	    (set! mask (fix:+ mask (C-enum "GDK_HINT_WIN_GRAVITY"))))))

    (C-call "gtk_window_set_geometry_hints"
	    (gobject-alien window)
	    (gobject-alien widget)
	    geometry mask)
    (free geometry)))

(define (gtk-window-resize window width height)
  (assert-glib-locked 'gtk-window-resize)
  (guarantee-gtk-window window 'gtk-window-resize)
  (guarantee positive-fixnum? width 'gtk-window-resize)
  (guarantee positive-fixnum? height 'gtk-window-resize)
  (C-call "gtk_window_resize" (gobject-alien window) width height))

(define (gtk-window-present window)
  (assert-glib-locked 'gtk-window-present)
  (guarantee-gtk-window window 'gtk-window-present)
  (C-call "gtk_window_present" (gobject-alien window)))

(define (set-gtk-window-delete-event-callback! window callback)
  (guarantee-gtk-window window 'set-gtk-window-delete-event-callback!)
  (guarantee-procedure-of-arity callback 1 'set-gtk-window-delete-event-callback!)
  (g-signal-connect window (C-callback "delete_event")
		    (make-delete-event-callback callback)))

(define (make-delete-event-callback callback)
  (named-lambda (delete-event-callback window GdkEvent)
    (declare (ignore GdkEvent))
    (callback window)))

(define gtk-clipboard-timeout 5000)

(define (gtk-window-get-clipboard-text window)
  (assert-glib-locked 'gtk-window-get-clipboard-text)
  (guarantee-gtk-window window 'gtk-window-get-clipboard-text)
  (let* ((gdkdisplay (C-call "gtk_widget_get_display"
			     (make-alien '|GtkDisplay|)
			     (gobject-alien window)))
	 (gdk-display (get-gdk-display gdkdisplay)))
    (gdk-display-get-clipboard-text gdk-display gtk-clipboard-timeout)))

(define (gtk-window-set-clipboard-text window string)
  (assert-glib-locked 'gtk-window-set-clipboard-text)
  (guarantee-gtk-window window 'gtk-window-set-clipboard-text)
  (guarantee string? string 'gtk-window-set-clipboard-text)
  (let* ((gdkdisplay (C-call "gtk_widget_get_display"
			    (make-alien '|GtkDisplay|)
			    (gobject-alien window)))
	 (gdk-display (get-gdk-display gdkdisplay)))
    (gdk-display-set-clipboard-text gdk-display string)))

(define %trace? #f)

(define-syntax %trace
  (syntax-rules ()
    ((_ ARGS ...)
     (if %trace? (outf-error ARGS ...)))))