#| -*-Scheme-*-

Copyright (C) 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015,
    2016, 2017, 2018, 2019 Matthew Birkholz

This file is part of a Gtk plugin for MIT/GNU Scheme Pucked.

This plugin is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2 of the License, or (at your
option) any later version.

This plugin is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this plugin; if not, write to the Free Software Foundation,
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

|#

;;;; Core utilities.
;;; package: (gtk)

(define c-includes (load-c-includes "gtk"))

(define-syntax define-integrable-operator
  #;(er-macro-transformer
   (lambda (form rename compare)
     (declare (ignore compare))
     (cond ((syntax-match? '((IDENTIFIER . MIT-BVL) + FORM) (cdr form))
	    (let ((r-begin (rename 'BEGIN))
		  (r-declare (rename 'DECLARE))
		  (r-define (rename 'DEFINE)))
	      `(,r-begin
		(,r-declare (INTEGRATE-OPERATOR ,(caadr form)))
		(,r-define ,@(cdr form)))))
	   (else
	    (ill-formed-syntax form)))))

  (rsc-macro-transformer
   (lambda (form environment)
     (declare (ignore environment))
     (if (syntax-match? '((IDENTIFIER . MIT-BVL) + FORM) (cdr form))
	 `(BEGIN
	   (DECLARE (INTEGRATE-OPERATOR ,(caadr form)))
	   (DEFINE ,@(cdr form)))
	 (ill-formed-syntax form)))))

(define-syntax error-if-null
  (syntax-rules ()
    ((_ ALIEN MESSAGE ...)
     (if (alien-null? ALIEN)
	 (error MESSAGE ...)))))

(define-integrable-operator (fix:max n m) (if (fix:> n m) n m))

(define-integrable-operator (fix:min n m) (if (fix:< n m) n m))

(define-integrable-operator (fix:negate i) (fix:- 0 i))

(define-integrable-operator (fix:abs n)
  (if (fix:negative? n) (fix:negate n) n))

(define-integrable (bit-mask-indices num)
  ;; The indices of the bits set in NUM.
  (let ((str (unsigned-integer->bit-string 32 num)))
    (let loop ((start 0))
      (let ((next (bit-substring-find-next-set-bit str start 32)))
	(if next
	    (cons next (loop (fix:1+ next)))
	    '())))))

(define-integrable (bit? int mask)
  (not (int:zero? (bitwise-and int mask))))

(declare (integrate-operator bit-ior))
(define (bit-ior . ints)
  (reduce bitwise-ior 0 ints))

(define (color? object)
  (and (flo:flonum? object) (fix:= 4 (flo:vector-length object))))

(define (make-rgba-color red green blue alpha)
  (let ((color (flo:vector-cons 4)))
    (flo:vector-set! color 0 red)
    (flo:vector-set! color 1 green)
    (flo:vector-set! color 2 blue)
    (flo:vector-set! color 3 alpha)
    color))