/* -*-C-*-

Copyright (C) 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015,
    2016, 2017, 2018, 2019 Matthew Birkholz

This file is part of a Gtk plugin for MIT/GNU Scheme Pucked.

This plugin is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2 of the License, or (at your
option) any later version.

This plugin is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this plugin; if not, write to the Free Software Foundation,
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

*/

/* A slice hook that updates a GtkWindow with the slice count and
   channels in the select registry. */

#include <mit-scheme.h>
#include <gtk/gtk.h>
/* #include <glib.h> */
/* #include <math.h> */
#include <stdlib.h>

static GtkWidget * slice_window = NULL;
static GtkWidget * slice_label;
static GtkWidget * status_label;
static void open_slice_window (void);
static void close_slice_window (void);
static gboolean slice_window_delete_event (GtkWidget *window, GdkEvent *event, gpointer *data);

extern void (*slice_hook)(void);
extern int slice_counter;
extern gchar * current_gpollfds_string (void);

void
gtk_slice_hook (void)
{
  /* Update the time-slice window before "sleeping". */
  if (slice_window != NULL)
    {
      gchar * fdstr, * text;

      text = g_strdup_printf ("Scheme time-slice: %d", slice_counter);
      gtk_label_set_text (GTK_LABEL (slice_label), text);
      g_free (text);
      fdstr = current_gpollfds_string ();
      text = g_strdup_printf ("Channels:%s", fdstr);
      if (fdstr[0] != '\0')
	g_free (fdstr);
      gtk_label_set_text (GTK_LABEL (status_label), text);
      g_free (text);
    }

  /* Force expose event delivery, so that animations continue to move
     even when Scheme is not "idle". */
  gdk_window_process_all_updates ();
}

static void
open_slice_window (void)
{
  slice_window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  GtkWidget * grid = gtk_grid_new ();
  status_label  = gtk_label_new ("Channels: -");
  slice_label  = gtk_label_new ("Scheme time-slice: 0");
  g_signal_connect (slice_window, "delete_event",
		    G_CALLBACK (slice_window_delete_event), NULL);
  gtk_grid_set_row_spacing (GTK_GRID (grid), 5);
  gtk_orientable_set_orientation (GTK_ORIENTABLE (grid),
				  GTK_ORIENTATION_VERTICAL);
  gtk_container_add (GTK_CONTAINER (grid), status_label);
  gtk_container_add (GTK_CONTAINER (grid), slice_label);
  gtk_container_add (GTK_CONTAINER (slice_window), grid);
  gtk_container_set_border_width (GTK_CONTAINER (slice_window), 5);
  gtk_window_set_title (GTK_WINDOW (slice_window), "Scheme Time-Slice Counter");
  gtk_window_set_type_hint (GTK_WINDOW (slice_window),
			    GDK_WINDOW_TYPE_HINT_UTILITY);
  gtk_widget_show_all (slice_window);
  /* gtk_window_parse_geometry (GTK_WINDOW (slice_window), "-0-0"); */
  slice_hook = &gtk_slice_hook;
}

static void
close_slice_window (void)
{
  slice_hook = NULL;
  gtk_widget_destroy (GTK_WIDGET (slice_window));
  slice_window = NULL;
  gtk_widget_destroy (GTK_WIDGET (status_label));
  status_label = NULL;
  gtk_widget_destroy (GTK_WIDGET (slice_label));
  slice_label = NULL;
}

static gboolean
slice_window_delete_event (GtkWidget *window, GdkEvent *event, gpointer *data)
{
  slice_window = NULL;
  gtk_widget_destroy (GTK_WIDGET (status_label));
  status_label = NULL;
  gtk_widget_destroy (GTK_WIDGET (slice_label));
  slice_label = NULL;
  return (FALSE);		/* not handled here */
}

gboolean
gtk_time_slice_window_p (void)
{
  return (slice_window == NULL ? 0 : 1);
}

void
gtk_time_slice_window (gboolean open_p)
{
  if (!open_p)
    {
      if (slice_window != NULL)
	close_slice_window ();
    }
  else
    {
      if (slice_window == NULL)
	open_slice_window ();
    }
}
