#| -*-Scheme-*-

Copyright (C) 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015,
    2016, 2017, 2018, 2019 Matthew Birkholz

This file is part of a Gtk plugin for MIT/GNU Scheme Pucked.

This plugin is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2 of the License, or (at your
option) any later version.

This plugin is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this plugin; if not, write to the Free Software Foundation,
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

|#

;;;; Initialize per $DISPLAY.
;;; package: (gtk main)

(define initialized?)

(define (gtk-initialized?)
  initialized?)

(define (gtk-start)
  ;; Called from gtk/make.scm, from a (load-option 'Gtk).
  (init-gtk ((ucode-primitive scheme-program-name 0))
	    (command-line-arguments)))

(define (initialize-package!)
  (reset-gtk!)
  (add-event-receiver! event:after-restore reset-gtk!)
  (restart-gtk!)
  (add-event-receiver! event:after-restart restart-gtk!))

(define (restart-gtk!)
  (let* ((program-name ((ucode-primitive scheme-program-name 0)))
	 (val (ignore-errors
	       (lambda () (init-gtk program-name (command-line-arguments))))))
    (if (condition? val)
	(warn val)
	(set! *command-line-arguments* val))))

(define (reset-gtk!)
  (for-each (lambda (w)
	      (alien-null! (gobject-alien w)))
	    toplevel-windows)
  (set! toplevel-windows '())
  (set! initialized? #f))

(define (init-gtk name args)
  (if (let ((s (get-environment-variable "DISPLAY")))
	(and (string? s) (not (string-null? s))))
      (init-gtk* name args)
      (warn "DISPLAY not set")))

(define (init-gtk* name args)
  ;; Call gtk_init_check.  Warn if it returns 0.  Return a list of
  ;; unused ARGS.
  (let ((arg-count (guarantee-list-of-type->length
		    args string? "list of commandline arguments (strings)"
		    'INIT-GTK))
	(vars-size (+ (C-sizeof "int")		;gtk_init_check return var
		      (C-sizeof "* * char"))))	;gtk_init_check return var
    (guarantee string? name 'INIT-GTK)
    (let* ((words (cons name args))
	   (words-bv (map string->utf8 words))
	   (vector-size
	    (* (C-sizeof "* char") (+ 1 arg-count)))
	   (total-size
	    (+ vars-size vector-size
	       (fold-left (lambda (sum arg)
			    (+ sum (bytevector-length arg) 1)) ;null terminated
			  0 words-bv)))
	   (bytes (malloc total-size #f))
	   (vector (alien-byte-increment bytes vars-size))
	   (word-scan (alien-byte-increment vector vector-size))
	   (vector-scan (copy-alien vector))
	   (count-var bytes)
	   (vector-var (alien-byte-increment count-var (C-sizeof "int"))))
      (for-each (lambda (word-bv)
		  (c-poke-pointer! vector-scan word-scan)
		  (c-poke-string! word-scan word-bv))
		words-bv)
      (C->= count-var "int" (+ 1 arg-count))
      (C->= vector-var "* * char" vector)
      (if (fix:zero? (with-glib-lock
		      (lambda ()
			(C-call "gtk_init_check" count-var vector-var))))
	  (warn "Could not initialize Gtk.")
	  (let ((new-argc (C-> count-var "int")))
	    (C-> vector-var "* * char" vector-scan)
	    (let ((new-args
		   (let loop ((i 0)(args '()))
		     (if (fix:< i new-argc)
			 (loop (fix:1+ i)
			       (cons (c-peek-cstringp! vector-scan)
				     args))
			 (reverse! args)))))
	      (free bytes)
	      (set! initialized? #t)
	      (cdr new-args)))))))

(define (gtk-time-slice-window?)
  (C-call "gtk_time_slice_window_p"))

(define (gtk-time-slice-window! open?)
  (with-glib-lock (lambda () (C-call "gtk_time_slice_window" (if open? 1 0)))))

(initialize-package!)